import asyncio
import concurrent.futures
import json
import logging
import multiprocessing
import socket
from typing import Dict, List, Optional, Union

import common.errors as errors
import common.util as utils
import uvicorn
from common.logging import setup_logging
from common.serialization import (
    DeepNumpyEncoder,
    truss_msgpack_deserialize,
    truss_msgpack_serialize,
)
from fastapi import Depends, FastAPI, Request
from fastapi.responses import ORJSONResponse
from fastapi.routing import APIRoute as FastAPIRoute
from model_wrapper import ModelWrapper
from starlette.responses import Response


async def parse_body(request: Request) -> bytes:
    """
    Used by FastAPI to read body in an asynchronous manner
    """
    return await request.body()


FORMAT = "%(asctime)s.%(msecs)03d %(name)s %(levelname)s [%(funcName)s():%(lineno)s] %(message)s"
DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
logging.basicConfig(level=logging.INFO, format=FORMAT, datefmt=DATE_FORMAT)


class UvicornCustomServer(multiprocessing.Process):
    def __init__(
        self, config: uvicorn.Config, sockets: Optional[List[socket.socket]] = None
    ):
        super().__init__()
        self.sockets = sockets
        self.config = config

    def stop(self):
        self.terminate()

    def run(self):
        server = uvicorn.Server(config=self.config)
        asyncio.run(server.serve(sockets=self.sockets))


class BasetenEndpoints:
    """The implementation of the model server endpoints.

    Historically, we relied on the kserve server interface, which assumes that
    multiple models are running behind a registry. As a result, some arguments to
    to functions will rename unused except for backwards compatibility checks.
    """

    def __init__(self, model: ModelWrapper) -> None:
        self._model = model

    def _safe_lookup_model(self, model_name: str) -> ModelWrapper:
        if model_name != self._model.name:
            raise errors.ModelMissingError(model_name)
        return self._model

    @staticmethod
    def check_healthy(model: ModelWrapper):
        if model.load_failed():
            raise errors.InferenceError("Model load failed")

        if not model.ready:
            raise errors.ModelNotReady(model.name)

    async def model_ready(self, model_name: str) -> Dict[str, Union[str, bool]]:
        self.check_healthy(self._safe_lookup_model(model_name))

        return {}

    async def invocations_ready(self) -> Dict[str, Union[str, bool]]:
        """
        This method provides compatibility with Sagemaker hosting for the 'ping' endpoint.
        """
        if self._model is None:
            raise errors.ModelMissingError("model")
        self.check_healthy(self._model)

        return {}

    def invocations(
        self, request: Request, body_raw: bytes = Depends(parse_body)
    ) -> Response:
        """
        This method provides compatibility with Sagemaker hosting for the 'invocations' endpoint.
        """
        return self.predict(self._model.name, request, body_raw)

    def predict(
        self, model_name: str, request: Request, body_raw: bytes = Depends(parse_body)
    ) -> Response:
        """
        This method is called by FastAPI, which introspects that it's not async, and schedules it on a thread
        """
        model: ModelWrapper = self._safe_lookup_model(model_name)

        self.check_healthy(model)

        body: Dict
        if self.is_binary(request):
            body = truss_msgpack_deserialize(body_raw)
        else:
            body = json.loads(body_raw)

        # calls ModelWrapper.__call__, which runs validate, preprocess, predict, and postprocess
        response: Dict = asyncio.run(model(body, headers=dict(request.headers.items())))

        response_headers = {}
        if self.is_binary(request):
            response_headers["Content-Type"] = "application/octet-stream"
            return Response(
                content=truss_msgpack_serialize(response), headers=response_headers
            )
        else:
            response_headers["Content-Type"] = "application/json"
            return Response(
                content=json.dumps(response, cls=DeepNumpyEncoder),
                headers=response_headers,
            )

    @staticmethod
    def is_binary(request: Request):
        return (
            "Content-Type" in request.headers
            and request.headers["Content-Type"] == "application/octet-stream"
        )


class TrussServer:
    def __init__(self, http_port: int, config: Dict):
        self.http_port = http_port
        self._config = config
        self._model = ModelWrapper(self._config)
        self._endpoints = BasetenEndpoints(self._model)

    def on_startup(self):
        """
        This method will be started inside the main process, so here is where we want to setup our logging and model
        """
        setup_logging()

        self._model.start_load()

    def create_application(self):
        return FastAPI(
            title="Baseten Inference Server",
            docs_url=None,
            redoc_url=None,
            default_response_class=ORJSONResponse,
            on_startup=[self.on_startup],
            routes=[
                # liveness endpoint
                FastAPIRoute(r"/", lambda: True),
                # readiness endpoint
                FastAPIRoute(
                    r"/v1/models/{model_name}", self._endpoints.model_ready, tags=["V1"]
                ),
                FastAPIRoute(
                    r"/v1/models/{model_name}:predict",
                    self._endpoints.predict,
                    methods=["POST"],
                    tags=["V1"],
                ),
                FastAPIRoute(
                    r"/v1/models/{model_name}:predict_binary",
                    self._endpoints.predict,
                    methods=["POST"],
                    tags=["V1"],
                ),
                # Endpoint aliases for Sagemaker hosting
                FastAPIRoute(r"/ping", self._endpoints.invocations_ready),
                FastAPIRoute(
                    r"/invocations",
                    self._endpoints.invocations,
                    methods=["POST"],
                ),
            ],
            exception_handlers={
                errors.InferenceError: errors.inference_error_handler,
                errors.ModelNotFound: errors.model_not_found_handler,
                errors.ModelNotReady: errors.model_not_ready_handler,
                NotImplementedError: errors.not_implemented_error_handler,
                Exception: errors.generic_exception_handler,
            },
        )

    def start(self):
        cfg = uvicorn.Config(
            self.create_application(),
            host="0.0.0.0",
            port=self.http_port,
            workers=1,
            log_config={
                "version": 1,
                "formatters": {
                    "default": {
                        "()": "uvicorn.logging.DefaultFormatter",
                        "datefmt": DATE_FORMAT,
                        "fmt": "%(asctime)s.%(msecs)03d %(name)s %(levelprefix)s %(message)s",
                        "use_colors": None,
                    },
                    "access": {
                        "()": "uvicorn.logging.AccessFormatter",
                        "datefmt": DATE_FORMAT,
                        "fmt": "%(asctime)s.%(msecs)03d %(name)s %(levelprefix)s %(client_addr)s %(process)s - "
                        '"%(request_line)s" %(status_code)s',
                        # noqa: E501
                    },
                },
                "handlers": {
                    "default": {
                        "formatter": "default",
                        "class": "logging.StreamHandler",
                        "stream": "ext://sys.stderr",
                    },
                    "access": {
                        "formatter": "access",
                        "class": "logging.StreamHandler",
                        "stream": "ext://sys.stdout",
                    },
                },
                "loggers": {
                    "uvicorn": {"handlers": ["default"], "level": "INFO"},
                    "uvicorn.error": {"level": "INFO"},
                    "uvicorn.access": {
                        "handlers": ["access"],
                        "level": "INFO",
                        "propagate": False,
                    },
                },
            },
        )

        max_asyncio_workers = min(32, utils.cpu_count() + 4)
        logging.info(f"Setting max asyncio worker threads as {max_asyncio_workers}")
        asyncio.get_event_loop().set_default_executor(
            concurrent.futures.ThreadPoolExecutor(max_workers=max_asyncio_workers)
        )

        async def serve():
            serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            serversocket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            serversocket.bind((cfg.host, cfg.port))
            serversocket.listen(5)

            logging.info(f"starting uvicorn with {cfg.workers} workers")
            for _ in range(cfg.workers):
                server = UvicornCustomServer(config=cfg, sockets=[serversocket])
                server.start()

        async def servers_task():
            servers = [serve()]
            await asyncio.gather(*servers)

        asyncio.run(servers_task())



import os
from contextlib import contextmanager


@contextmanager
def current_directory(directory: str):
    """
    Execute with using given directory as current working directory.

    This can be problematic in a multi-threaded scenario. It's assumed that
    the caller will take locks as needed.
    """
    cwd = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(cwd)



from typing import Callable, Optional, TypeVar

X = TypeVar("X")
Y = TypeVar("Y")


def transform_optional(x: Optional[X], fn: Callable[[X], Optional[Y]]) -> Optional[Y]:
    if x is None:
        return None

    return fn(x)



import tempfile
from pathlib import Path

import pytest
import requests
from python_on_whales import docker
from truss.constants import CONFIG_FILE, TRUSS
from truss.contexts.image_builder.serving_image_builder import (
    ServingImageBuilderContext,
)
from truss.contexts.local_loader.load_model_local import LoadModelLocal
from truss.model_frameworks.huggingface_transformer import HuggingfaceTransformer
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.truss_config import TrussConfig
from truss.truss_handle import wait_for_truss


def test_to_truss(huggingface_transformer_t5_small_pipeline):
    with tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        model = huggingface_transformer_t5_small_pipeline
        truss_dir = Path(tmp_work_dir, "truss")
        framework = HuggingfaceTransformer()
        framework.to_truss(model, truss_dir)

        # Assertions
        config = TrussConfig.from_yaml(truss_dir / CONFIG_FILE)
        assert config.model_class_filename == "model.py"
        assert config.model_class_name == "Model"
        assert config.model_framework == framework.typ()
        assert config.model_type == "text2text-generation"
        assert len(config.requirements) == 0

        assert config.python_version.startswith("py3")
        model_metadata = config.model_metadata
        assert not model_metadata["has_named_args"]
        assert not model_metadata["has_hybrid_args"]

        assert (truss_dir / "data" / "model").exists()
        assert (truss_dir / "model" / "model.py").exists()


def test_run_truss(huggingface_transformer_t5_small_pipeline):
    with tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        model = huggingface_transformer_t5_small_pipeline
        truss_dir = Path(tmp_work_dir, "truss")
        framework = HuggingfaceTransformer()
        framework.to_truss(model, truss_dir)
        model = LoadModelLocal.run(truss_dir)
        result = model.predict("My name is Sarah and I live in London")
        predictions = result["predictions"]
        assert len(predictions) == 1
        prediction = predictions[0]
        assert prediction["generated_text"].startswith("Mein Name")


@pytest.mark.integration
def test_run_image(huggingface_transformer_t5_small_pipeline):
    with ensure_kill_all(), tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        model = huggingface_transformer_t5_small_pipeline
        truss_dir = Path(tmp_work_dir, "truss")
        framework = HuggingfaceTransformer()
        framework.to_truss(model, truss_dir)
        tag = f"test-{framework.typ().value}-model:latest"
        image = ServingImageBuilderContext.run(truss_dir).build_image(
            tag=tag, labels={TRUSS: True}
        )
        assert image.repo_tags == [tag]
        container = docker.run(tag, publish=[[8080, 8080]], detach=True)

        wait_for_truss("http://localhost:8080/v1/models/model", container)

        resp = requests.post(
            "http://localhost:8080/v1/models/model:predict",
            json="My name is Sarah and I live in London",
        )

        predictions = resp.json()["predictions"]
        assert len(predictions) == 1
        prediction = predictions[0]
        assert prediction["generated_text"].startswith("Mein Name")



import importlib
import inspect
import logging
import os
import sys
import time
import traceback
from enum import Enum
from pathlib import Path
from threading import Lock, Thread
from typing import Any, Dict, Optional

from common.patches import apply_patches
from common.retry import retry
from shared.secrets_resolver import SecretsResolver

MODEL_BASENAME = "model"

NUM_LOAD_RETRIES = int(os.environ.get("NUM_LOAD_RETRIES_TRUSS", "3"))


class ModelWrapper:
    class Status(Enum):
        NOT_READY = 0
        LOADING = 1
        READY = 2
        FAILED = 3

    def __init__(self, config: Dict):
        self._config = config
        self._logger = logging.getLogger()
        self.name = MODEL_BASENAME
        self.ready = False
        self._load_lock = Lock()
        self._predict_lock = Lock()
        self._status = ModelWrapper.Status.NOT_READY

    def load(self) -> bool:
        if self.ready:
            return self.ready

        # if we are already loading, just pass; our container will return 503 while we're loading
        if not self._load_lock.acquire(blocking=False):
            return False

        self._status = ModelWrapper.Status.LOADING

        self._logger.info("Executing model.load()...")

        try:
            start_time = time.perf_counter()
            self.try_load()
            self.ready = True
            self._status = ModelWrapper.Status.READY
            self._logger.info(
                f"Completed model.load() execution in {_elapsed_ms(start_time)} ms"
            )

            return self.ready
        except Exception:
            self._logger.exception("Exception while loading model")
            self._status = ModelWrapper.Status.FAILED
        finally:
            self._load_lock.release()

        return self.ready

    def start_load(self):
        if self.should_load():
            thread = Thread(target=self.load)
            thread.start()

    def load_failed(self) -> bool:
        return self._status == ModelWrapper.Status.FAILED

    def should_load(self) -> bool:
        # don't retry failed loads
        return (
            not self._load_lock.locked()
            and not self._status == ModelWrapper.Status.FAILED
            and not self.ready
        )

    def try_load(self):
        data_dir = Path("data")
        data_dir.mkdir(exist_ok=True)

        if "bundled_packages_dir" in self._config:
            bundled_packages_path = Path("/packages")
            if bundled_packages_path.exists():
                sys.path.append(str(bundled_packages_path))
        model_module_name = str(
            Path(self._config["model_class_filename"]).with_suffix("")
        )
        module = importlib.import_module(
            f"{self._config['model_module_dir']}.{model_module_name}"
        )
        model_class = getattr(module, self._config["model_class_name"])
        model_class_signature = inspect.signature(model_class)
        model_init_params = {}
        if _signature_accepts_keyword_arg(model_class_signature, "config"):
            model_init_params["config"] = self._config
        if _signature_accepts_keyword_arg(model_class_signature, "data_dir"):
            model_init_params["data_dir"] = data_dir
        if _signature_accepts_keyword_arg(model_class_signature, "secrets"):
            model_init_params["secrets"] = SecretsResolver.get_secrets(self._config)
        apply_patches(
            self._config.get("apply_library_patches", True),
            self._config["requirements"],
        )
        self._model = model_class(**model_init_params)

        if hasattr(self._model, "load"):
            retry(
                self._model.load,
                NUM_LOAD_RETRIES,
                self._logger.warn,
                "Failed to load model.",
                gap_seconds=1.0,
            )

    def preprocess(
        self,
        payload: Any,
        headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        if not hasattr(self._model, "preprocess"):
            return payload
        return self._model.preprocess(payload)  # type: ignore

    def postprocess(
        self,
        response: Any,
        headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        if not hasattr(self._model, "postprocess"):
            return response
        return self._model.postprocess(response)  # type: ignore

    def predict(
        self,
        payload: Any,
        headers: Optional[Dict[str, str]] = None,
    ) -> Any:
        try:
            self._predict_lock.acquire()
            return self._model.predict(payload)  # type: ignore
        except Exception:
            response = {}
            logging.exception("Exception while running predict")
            response["error"] = {"traceback": traceback.format_exc()}
            return response
        finally:
            self._predict_lock.release()

    async def __call__(
        self, body: Any, headers: Optional[Dict[str, str]] = None
    ) -> Dict:
        """Method to call predictor or explainer with the given input.

        Args:
            body (Any): Request payload body.
            headers (Dict): Request headers.

        Returns:
            Dict: Response output from preprocess -> predictor -> postprocess
        """

        payload = (
            await self.preprocess(body, headers)
            if inspect.iscoroutinefunction(self.preprocess)
            else self.preprocess(body, headers)
        )

        response = (
            (await self.predict(payload, headers))
            if inspect.iscoroutinefunction(self.predict)
            else self.predict(payload, headers)
        )

        response = self.postprocess(response, headers)

        return response


def _signature_accepts_keyword_arg(signature: inspect.Signature, kwarg: str) -> bool:
    return kwarg in signature.parameters or _signature_accepts_kwargs(signature)


def _signature_accepts_kwargs(signature: inspect.Signature) -> bool:
    for param in signature.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            return True
    return False


def _elapsed_ms(since_micro_seconds: float) -> int:
    return int((time.perf_counter() - since_micro_seconds) * 1000)



import os
import sys

import psutil


def model_supports_predict_proba(model: object) -> bool:
    if not hasattr(model, "predict_proba"):
        return False
    if hasattr(
        model, "_check_proba"
    ):  # noqa eg Support Vector Machines *can* predict proba if they made certain choices while training
        try:
            model._check_proba()
            return True
        except AttributeError:
            return False
    return True


def cpu_count():
    """Get the available CPU count for this system.
    Takes the minimum value from the following locations:
    - Total system cpus available on the host.
    - CPU Affinity (if set)
    - Cgroups limit (if set)
    """
    count = os.cpu_count()

    # Check CPU affinity if available
    try:
        affinity_count = len(psutil.Process().cpu_affinity())
        if affinity_count > 0:
            count = min(count, affinity_count)
    except Exception:
        pass

    # Check cgroups if available
    if sys.platform == "linux":
        try:
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_quota_us") as f:
                quota = int(f.read())
            with open("/sys/fs/cgroup/cpu,cpuacct/cpu.cfs_period_us") as f:
                period = int(f.read())
            cgroups_count = int(quota / period)
            if cgroups_count > 0:
                count = min(count, cgroups_count)
        except Exception:
            pass

    return count



import pytest
import torch
from torch import package
from truss.model_frameworks.pytorch import (
    TORCH_MODEL_PACKAGE_NAME,
    TORCH_MODEL_PICKLE_FILENAME,
    TORCH_PACKAGE_FILE,
    PyTorch,
)
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.truss_handle import TrussHandle


def test_serialize_model_to_directory(pytorch_model_with_numpy_import, tmp_path):
    pytorch = PyTorch()
    model = pytorch_model_with_numpy_import[0]
    pytorch.serialize_model_to_directory(model, tmp_path)
    pkg_file_path = tmp_path / TORCH_PACKAGE_FILE
    assert pkg_file_path.exists()
    imp = package.PackageImporter(pkg_file_path)
    model = imp.load_pickle(TORCH_MODEL_PACKAGE_NAME, TORCH_MODEL_PICKLE_FILENAME)
    result = model(torch.tensor([[0, 0, 0]], dtype=torch.float32)).tolist()
    assert len(result) == 1


def test_supports_model_class(pytorch_model_with_numpy_import):
    pytorch = PyTorch()
    assert pytorch.supports_model_class(pytorch_model_with_numpy_import[0].__class__)


@pytest.mark.integration
def test_run_image(pytorch_model_with_numpy_import, tmp_path):
    truss_dir = tmp_path / "truss"
    pytorch = PyTorch()
    model = pytorch_model_with_numpy_import[0]
    pytorch.to_truss(model, truss_dir)
    truss = TrussHandle(truss_dir)
    with ensure_kill_all():
        result = truss.docker_predict(
            [[0, 0, 0]],
            local_port=8090,
        )
        assert len(result["predictions"]) == 1



from typing import Any, Dict

import requests
from flask import Blueprint, Response, current_app, jsonify, request
from helpers.errors import ModelNotReady
from requests.exceptions import ConnectionError
from tenacity import Retrying, retry_if_exception_type, stop_after_attempt, wait_fixed

INFERENCE_SERVER_START_WAIT_SECS = 60


control_app = Blueprint("control", __name__)


@control_app.route("/")
def index():
    return jsonify({})


@control_app.route("/v1/<path:path>", methods=["GET", "POST"])
def proxy(path):
    inference_server_port = current_app.config["inference_server_port"]
    inference_server_process_controller = current_app.config[
        "inference_server_process_controller"
    ]

    # Wait a bit for inference server to start
    for attempt in Retrying(
        retry=(
            retry_if_exception_type(ConnectionError)
            | retry_if_exception_type(ModelNotReady)
        ),
        stop=stop_after_attempt(INFERENCE_SERVER_START_WAIT_SECS),
        wait=wait_fixed(1),
    ):
        with attempt:
            try:
                resp = requests.request(
                    method=request.method,
                    url=f"http://localhost:{inference_server_port}/v1/{path}",
                    data=request.get_data(),
                    cookies=request.cookies,
                    headers=request.headers,
                )
                if _is_model_not_ready(resp):
                    raise ModelNotReady("Model has started running, but not ready yet.")
            except ConnectionError as exp:
                # This check is a bit expensive so we don't do it before every request, we
                # do it only if request fails with connection error. If the inference server
                # process is running then we continue waiting for it to start (by retrying),
                # otherwise we bail.
                if (
                    inference_server_process_controller.inference_server_ever_started()
                    and not inference_server_process_controller.is_inference_server_running()
                ):
                    error_msg = "It appears your model has stopped running. This often means' \
                        ' it crashed and may need a fix to get it running again."
                    return Response(error_msg, 503)
                raise exp

    headers = [(name, value) for (name, value) in resp.raw.headers.items()]
    response = Response(resp.content, resp.status_code, headers)
    return response


@control_app.route("/control/patch", methods=["POST"])
def patch() -> Dict[str, str]:
    current_app.logger.info("Patch request received.")
    patch_request = request.get_json()
    current_app.config["inference_server_controller"].apply_patch(patch_request)
    current_app.logger.info("Patch applied successfully")
    return {"msg": "Patch applied successfully"}


@control_app.route("/control/truss_hash", methods=["GET"])
def truss_hash() -> Dict[str, Any]:
    t_hash = current_app.config["inference_server_controller"].truss_hash()
    return {"result": t_hash}


@control_app.route("/control/restart_inference_server", methods=["POST"])
def restart_inference_server() -> Dict[str, str]:
    current_app.config["inference_server_controller"].restart()

    return {"msg": "Inference server started successfully"}


@control_app.route("/control/has_partially_applied_patch", methods=["GET"])
def has_partially_applied_patch() -> Dict[str, Any]:
    app_has_partially_applied_patch = current_app.config[
        "inference_server_controller"
    ].has_partially_applied_patch()
    return {"result": app_has_partially_applied_patch}


@control_app.route("/control/stop_inference_server", methods=["POST"])
def stop_inference_server() -> Dict[str, str]:
    current_app.config["inference_server_controller"].stop()
    return {"msg": "Inference server stopped successfully"}


def _is_model_not_ready(resp) -> bool:
    return (
        resp.status_code == 503
        and resp.content is not None
        and "model is not ready" in resp.content.decode("utf-8")
    )



import logging
import traceback
from typing import Any

import torch
from transformers import pipeline


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs["data_dir"]
        config = kwargs["config"]
        self._config = config
        model_metadata = config["model_metadata"]
        self._transformer_config = model_metadata["transformer_config"]
        self._has_named_args = model_metadata["has_named_args"]
        self._has_hybrid_args = model_metadata["has_hybrid_args"]
        self._model = None

    def load(self):
        transformer_config = self._transformer_config.copy()
        if torch.cuda.is_available():
            transformer_config["device"] = 0

        self._model = pipeline(
            task=self._config["model_type"],
            **transformer_config,
        )

    def predict(self, model_input: Any) -> Any:
        model_output = {}
        instances = model_input

        with torch.no_grad():
            if self._has_named_args:
                result = [self._model(**instance) for instance in instances]
            elif self._has_hybrid_args:
                try:
                    result = []
                    for instance in instances:
                        prompt = instance.pop("prompt")
                        result.append(self._model(prompt, **instance))
                except (KeyError, AttributeError):
                    logging.error(traceback.format_exc())
                    model_output["error"] = {
                        "traceback": f'Expected request as an object with text in "prompt"\n{traceback.format_exc()}'
                    }
                    return model_output
            else:
                result = self._model(instances)
        model_output["predictions"] = result
        return model_output



from pathlib import Path

from truss.patch.dir_signature import directory_content_signature
from truss.patch.types import TrussSignature


def calc_truss_signature(truss_dir: Path) -> TrussSignature:
    content_signature = directory_content_signature(truss_dir)
    with (truss_dir / "config.yaml").open("r") as config_file:
        config = config_file.read()
    return TrussSignature(
        content_hashes_by_path=content_signature,
        config=config,
    )



import logging
from pathlib import Path
from typing import Any, Dict, Optional, Set

from truss.constants import HUGGINGFACE_TRANSFORMER_MODULE_NAME
from truss.model_framework import ModelFramework
from truss.types import ModelFrameworkType

logger: logging.Logger = logging.getLogger(__name__)


class HuggingfaceTransformer(ModelFramework):
    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.HUGGINGFACE_TRANSFORMER

    def required_python_depedencies(self) -> Set[str]:
        return HUGGINGFACE_TRANSFORMER_MODULE_NAME

    def serialize_model_to_directory(self, model, target_directory: Path):
        # For Huggingface models, all the important details are in metadata.
        pass

    def model_metadata(self, model) -> Dict[str, Any]:
        hf_task = self.model_type(model)
        return {
            "transformer_config": {
                "model": _hf_model_name(model),
            },
            "has_hybrid_args": hf_task in {"text-generation"},
            "has_named_args": hf_task in {"zero-shot-classification"},
        }

    def model_type(self, model) -> str:
        return _infer_hf_task(model)

    def supports_model_class(self, model_class) -> bool:
        model_framework, _, _ = model_class.__module__.partition(".")
        return model_framework == "transformers"


def _hf_model_name(model) -> Optional[str]:
    try:
        return model.model.config._name_or_path  # type: ignore
    except AttributeError:
        logger.info(
            "Unable to infer a HuggingFace model on this task pipeline; transformer library will default"
        )
        return None


def _infer_hf_task(model) -> str:
    try:
        return model.task
    except AttributeError as error:
        logger.exception(
            "Unable to find a HuggingFace task on this object, did you call with a `Pipeline` object"
        )
        raise error



import contextlib
import importlib
import os
import subprocess
import sys
import time
from pathlib import Path

import lightgbm as lgb
import numpy as np
import pandas as pd
import pytest
import requests
import tensorflow as tf
from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from tensorflow.keras import layers
from tensorflow.keras.layers.experimental import preprocessing
from transformers import AutoModelWithLMHead, AutoTokenizer, pipeline
from truss.build import create, init
from truss.contexts.image_builder.serving_image_builder import (
    ServingImageBuilderContext,
)
from truss.contexts.local_loader.docker_build_emulator import DockerBuildEmulator
from truss.truss_config import DEFAULT_BUNDLED_PACKAGES_DIR
from truss.types import Example
from xgboost import XGBClassifier

PYTORCH_MODEL_FILE_CONTENTS = """
import torch
import torch.nn as nn

class MyModel(nn.Module):
    def __init__(self):
        super(MyModel, self).__init__()
        self.main = torch.nn.Sequential(
            torch.nn.Linear(3, 1),
            torch.nn.Flatten(0, 1)
        )

    def forward(self, input):
        return self.main(input)

"""

PYTORCH_MODEL_FILE_WITH_NUMPY_IMPORT_CONTENTS = """
import numpy as np
import torch
import torch.nn as nn

class MyModel(nn.Module):
    def __init__(self):
        super(MyModel, self).__init__()
        self.main = torch.nn.Sequential(
            torch.nn.Linear(3, 1),
            torch.nn.Flatten(0, 1)
        )

    def forward(self, input):
        return self.main(input)

"""


PYTORCH_WITH_INIT_ARGS_MODEL_FILE_CONTENTS = """
import torch
import torch.nn as nn

class MyModel(nn.Module):
    def __init__(self, arg1, arg2, kwarg1=1, kwarg2=2):
        super(MyModel, self).__init__()
        self.arg1 = arg1
        self.arg2 = arg2
        self.kwarg1 = kwarg1
        self.kwarg2 = kwarg2

    def forward(self, input):
        return input

"""


CUSTOM_MODEL_CODE = """
class Model:
    def __init__(*args, **kwargs):
        pass

    def load(self):
        pass

    def predict(self, model_input):
        return [1 for i in model_input]
"""

CUSTOM_MODEL_USING_EXTERNAL_PACKAGE_CODE = """
import top_module
import subdir.sub_module
import top_module2
class Model:
    def predict(self, model_input):
        return [1 for i in model_input]
"""

CUSTOM_MODEL_CODE_WITH_PRE_AND_POST_PROCESS = """
class Model:
    def __init__(*args, **kwargs):
        pass

    def load(*args, **kwargs):
        pass

    def preprocess(self, model_input):
        # Adds 1 to all
        return [value + 1 for value in model_input]

    def predict(self, model_input):
        # Returns inputs as predictions
        return {
            'predictions': model_input,
        }

    def postprocess(self, model_output):
        # Adds 2 to all
        return {
            'predictions': [value + 2 for value in model_output['predictions']],
        }
"""

CUSTOM_MODEL_CODE_USING_BUNDLED_PACKAGE = """
from test_package import test

class Model:
    def predict(self, request):
        # Returns inputs as predictions
        return {
            'predictions': [test.X],
        }

"""

CUSTOM_MODEL_CODE_FOR_GPU_TESTING = """
import subprocess

class Model:
    def predict(self, request):
        process = subprocess.run(['nvcc','--version'], check=True, stdout=subprocess.PIPE, universal_newlines=True)
        cuda_version = process.stdout.split('\\n')[-2].split()[-1].split('/')[0].split('_')[-1]
        return {
            'predictions': [{'cuda_version': cuda_version}],
        }
"""

CUSTOM_MODEL_CODE_FOR_SECRETS_TESTING = """
import subprocess

class Model:
    def __init__(self, secrets):
        self._secrets = secrets

    def predict(self, request):
        # Expects instance to be secret name and returns secret value as prediction
        secret_name = request['instances'][0]
        return {
            'predictions': [self._secrets[secret_name]],
        }
"""


# Doesn't implement load
NO_LOAD_CUSTOM_MODEL_CODE = """
class Model:
    def preprocess(self, request):
        return request

    def postprocess(self, request):
        return request

    def predict(self, request):
        return {
            'predictions': [1]
        }
"""


# Doesn't implement predict
NO_PREDICT_CUSTOM_MODEL_CODE = """
class MyModel:
    def load(self):
        pass
"""

LONG_LOAD_MODEL_CODE = """
import time
class Model:
     def load(*args, **kwargs):
        time.sleep(10)
        pass

     def predict(self, model_input):
        return {
            'predictions': model_input,
        }
"""

# Doesn't implement preprocess
NO_PREPROCESS_CUSTOM_MODEL_CODE = """
class Model:
     def load(*args, **kwargs):
        pass

     def postprocess(self, request):
        # Adds 1 to all
        return {
            'predictions': [value + 1 for value in request['predictions']],
        }

     def predict(self, model_input):
        return {
            'predictions': model_input,
        }
"""


# Doesn't implement postprocess
NO_POSTPROCESS_CUSTOM_MODEL_CODE = """
class Model:
     def load(*args, **kwargs):
        pass

     def preprocess(self, model_input):
        # Adds 1 to all
        return [value + 1 for value in model_input]

     def predict(self, model_input):
        return {
            'predictions': model_input,
        }
"""

# Implements no params for init
NO_PARAMS_INIT_CUSTOM_MODEL_CODE = """
class Model:
     def __init__(self):
        pass

     def preprocess(self, request):
        return request

     def postporcess(self, request):
        return request

     def predict(self, model_input):
        return {
            'predictions': model_input,
        }
"""

VARIABLES_TO_ARTIFACTS_TRAIN_CLASS_CODE = """
import json


class Train:
    def __init__(
       self,
       config,
       output_dir,
       variables,
    ):
        self.output_dir = output_dir
        self.variables = variables

    def train(self):
        with (self.output_dir / 'variables.json').open('w') as fp:
            fp.write(json.dumps(self.variables))

"""


EXTERNAL_DATA_ACCESS = """
class Model:
    def __init__(self, data_dir):
        self._data_dir = data_dir
        pass

    def predict(self, model_input):
        with (self._data_dir / 'test.txt').open() as file:
            return file.read()
"""


@pytest.fixture
def pytorch_model(tmp_path):
    return _pytorch_model_from_content(tmp_path, PYTORCH_MODEL_FILE_CONTENTS)


@pytest.fixture
def pytorch_model_with_numpy_import(tmp_path):
    return _pytorch_model_from_content(
        tmp_path, PYTORCH_MODEL_FILE_WITH_NUMPY_IMPORT_CONTENTS
    )


@pytest.fixture
def pytorch_model_init_args():
    return {"arg1": 1, "arg2": 2, "kwarg1": 3, "kwarg2": 4}


@pytest.fixture
def pytorch_model_with_init_args(tmp_path, pytorch_model_init_args):
    f = tmp_path / "my_model_with_init.py"
    f.write_text(PYTORCH_WITH_INIT_ARGS_MODEL_FILE_CONTENTS)

    sys.path.append(str(tmp_path))
    model_class = getattr(importlib.import_module("my_model_with_init"), "MyModel")
    return model_class(**pytorch_model_init_args), f


@pytest.fixture
def custom_model_truss_dir(tmp_path) -> Path:
    yield _custom_model_from_code(
        tmp_path,
        "custom_truss",
        CUSTOM_MODEL_CODE,
    )


@pytest.fixture
def no_preprocess_custom_model(tmp_path):
    yield _custom_model_from_code(
        tmp_path,
        "my_no_preprocess_model",
        NO_PREPROCESS_CUSTOM_MODEL_CODE,
    )


@pytest.fixture
def long_load_model(tmp_path):
    yield _custom_model_from_code(
        tmp_path,
        "long_load_model",
        LONG_LOAD_MODEL_CODE,
    )


@pytest.fixture
def custom_model_control(tmp_path):
    yield _custom_model_from_code(
        tmp_path,
        "control_truss",
        CUSTOM_MODEL_CODE,
        handle_ops=lambda handle: handle.live_reload(),
    )


@pytest.fixture
def custom_model_external_data_access_tuple_fixture(tmp_path: Path):
    content = "test"
    filename = "test.txt"
    (tmp_path / filename).write_text(content)
    port = 9089
    proc = subprocess.Popen(
        ["python", "-m", "http.server", str(port), "--bind", "*"],
        cwd=tmp_path,
    )
    try:
        url = f"http://localhost:{port}/{filename}"
        # Add arbitrary get params to get that they don't cause issues, the
        # server above ignores them.
        # url_with_get_params = f"{url}?foo=bar&baz=bla"
        url_with_get_params = f"{url}?foo=bar&baz=bla"
        yield (
            _custom_model_from_code(
                tmp_path,
                "external_data_access",
                EXTERNAL_DATA_ACCESS,
                handle_ops=lambda handle: handle.add_external_data_item(
                    url=url_with_get_params, local_data_path="test.txt"
                ),
            ),
            content,
        )
    finally:
        proc.kill()


@pytest.fixture
def custom_model_with_external_package(tmp_path: Path):
    ext_pkg_path = tmp_path / "ext_pkg"
    ext_pkg_path.mkdir()
    (ext_pkg_path / "subdir").mkdir()
    (ext_pkg_path / "subdir" / "sub_module.py").touch()
    (ext_pkg_path / "top_module.py").touch()
    ext_pkg_path2 = tmp_path / "ext_pkg2"
    ext_pkg_path2.mkdir()
    (ext_pkg_path2 / "top_module2.py").touch()

    def add_packages(handle):
        # Use absolute path for this
        handle.add_external_package(str(ext_pkg_path.resolve()))
        # Use relative path for this
        handle.add_external_package("../ext_pkg2")

    yield _custom_model_from_code(
        tmp_path,
        "control_truss",
        CUSTOM_MODEL_USING_EXTERNAL_PACKAGE_CODE,
        handle_ops=add_packages,
    )


@pytest.fixture
def no_postprocess_custom_model(tmp_path):
    yield _custom_model_from_code(
        tmp_path,
        "my_no_postprocess_model",
        NO_POSTPROCESS_CUSTOM_MODEL_CODE,
    )


@pytest.fixture
def no_load_custom_model(tmp_path):
    yield _custom_model_from_code(
        tmp_path,
        "my_no_load_model",
        NO_LOAD_CUSTOM_MODEL_CODE,
    )


@pytest.fixture
def no_params_init_custom_model(tmp_path):
    yield _custom_model_from_code(
        tmp_path,
        "my_no_params_init_load_model",
        NO_PARAMS_INIT_CUSTOM_MODEL_CODE,
    )


@pytest.fixture
def useless_file(tmp_path):
    f = tmp_path / "useless.py"
    f.write_text("")
    sys.path.append(str(tmp_path))
    return f


@contextlib.contextmanager
def temp_dir(directory):
    """A context to allow user to drop into the temporary
    directory created by the tmp_path fixture"""
    current_dir = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(current_dir)


@pytest.fixture(scope="session")
def xgboost_pima_model():
    pima_dataset_path = (
        Path(__file__).parent.parent / "test_data" / "pima-indians-diabetes.csv"
    )
    dataset = np.loadtxt(pima_dataset_path, delimiter=",")
    model = XGBClassifier()
    X = dataset[:, 0:8]
    Y = dataset[:, 8]
    model.fit(X, Y)
    return model


@pytest.fixture(scope="session")
def iris_dataset():
    return load_iris()


@pytest.fixture(scope="session")
def sklearn_rfc_model(iris_dataset):
    data_x = iris_dataset["data"]
    data_y = iris_dataset["target"]
    data_x = pd.DataFrame(data_x)
    rfc_model = RandomForestClassifier()
    rfc_model.fit(data_x, data_y)
    return rfc_model


@pytest.fixture(scope="session")
def lgb_pima_model():
    pima_dataset_path = (
        Path(__file__).parent.parent / "test_data" / "pima-indians-diabetes.csv"
    )
    params = {
        "boosting_type": "gbdt",
        "objective": "softmax",
        "metric": "multi_logloss",
        "num_leaves": 31,
        "num_classes": 2,
        "learning_rate": 0.05,
        "feature_fraction": 0.9,
        "bagging_fraction": 0.8,
        "bagging_freq": 5,
        "verbose": 0,
    }
    dataset = pd.read_csv(pima_dataset_path, header=None)
    Y = dataset[8]
    X = dataset.drop(8, axis=1)
    train = lgb.Dataset(X, Y)
    model = lgb.train(params=params, train_set=train)
    return model


@pytest.fixture(scope="session")
def mpg_dataset():
    # url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/auto-mpg/auto-mpg.data'
    auto_mpg_data = Path(__file__).parent.parent / "test_data" / "auto-mpg.data"
    column_names = [
        "MPG",
        "Cylinders",
        "Displacement",
        "Horsepower",
        "Weight",
        "Acceleration",
        "Model Year",
        "Origin",
    ]

    raw_dataset = pd.read_csv(
        str(auto_mpg_data),
        names=column_names,
        na_values="?",
        comment="\t",
        sep=" ",
        skipinitialspace=True,
    )

    dataset = raw_dataset.copy()
    dataset.isna().sum()
    dataset = dataset.dropna()
    dataset["Origin"] = dataset["Origin"].map({1: "USA", 2: "Europe", 3: "Japan"})
    dataset = pd.get_dummies(dataset, columns=["Origin"], prefix="", prefix_sep="")
    return dataset


@pytest.fixture(scope="session")
def keras_mpg_model(mpg_dataset):
    train_dataset = mpg_dataset.sample(frac=0.8, random_state=0)
    train_features = train_dataset.copy()
    train_labels = train_features.pop("MPG")
    normalizer = preprocessing.Normalization(
        input_shape=[
            9,
        ],
        axis=None,
    )
    normalizer.adapt(np.array(train_features))
    linear_model = tf.keras.Sequential([normalizer, layers.Dense(units=1)])

    linear_model.compile(
        optimizer=tf.optimizers.Adam(learning_rate=0.1), loss="mean_absolute_error"
    )

    linear_model.fit(
        train_features,
        train_labels,
        epochs=5,
        # suppress logging
        verbose=0,
        # Calculate validation results on 20% of the training data
        validation_split=0.2,
    )
    return linear_model


@pytest.fixture(scope="session")
def huggingface_transformer_t5_small_pipeline():
    tokenizer = AutoTokenizer.from_pretrained("t5-small")
    model = AutoModelWithLMHead.from_pretrained("t5-small")
    return pipeline("text2text-generation", model=model, tokenizer=tokenizer)


@pytest.fixture(scope="session")
def huggingface_transformer_t5_small_model():
    model = AutoModelWithLMHead.from_pretrained("t5-small")
    return model


@pytest.fixture(scope="session")
def huggingface_transformer_t5_small_tokenizer():
    tokenizer = AutoTokenizer.from_pretrained("t5-small")
    return tokenizer


@pytest.fixture
def custom_model_truss_dir_with_pre_and_post_no_example(tmp_path):
    dir_path = tmp_path / "custom_truss_with_pre_post_no_example"
    handle = init(str(dir_path))
    handle.spec.model_class_filepath.write_text(
        CUSTOM_MODEL_CODE_WITH_PRE_AND_POST_PROCESS
    )
    yield dir_path


@pytest.fixture
def huggingface_truss_handle_small_model(
    tmp_path: Path,
    huggingface_transformer_t5_small_pipeline,
):
    dir_path = tmp_path / "huggingface_truss_small_model"
    dir_path.mkdir()
    create(huggingface_transformer_t5_small_pipeline, target_directory=str(dir_path))
    return dir_path


@pytest.fixture
def custom_model_truss_dir_with_pre_and_post(tmp_path):
    dir_path = tmp_path / "custom_truss_with_pre_post"
    handle = init(str(dir_path))
    handle.spec.model_class_filepath.write_text(
        CUSTOM_MODEL_CODE_WITH_PRE_AND_POST_PROCESS
    )
    handle.update_examples([Example("example1", {"inputs": [[0]]})])
    yield dir_path


@pytest.fixture
def variables_to_artifacts_training_truss(tmp_path):
    dir_path = tmp_path / "training_truss"
    handle = init(str(dir_path), trainable=True)
    handle.spec.train_class_filepath.write_text(VARIABLES_TO_ARTIFACTS_TRAIN_CLASS_CODE)
    yield dir_path


@pytest.fixture
def custom_model_truss_dir_with_bundled_packages(tmp_path):
    truss_dir_path: Path = tmp_path / "custom_model_truss_dir_with_bundled_packages"
    handle = init(str(truss_dir_path))
    handle.spec.model_class_filepath.write_text(CUSTOM_MODEL_CODE_USING_BUNDLED_PACKAGE)
    packages_path = truss_dir_path / DEFAULT_BUNDLED_PACKAGES_DIR / "test_package"
    packages_path.mkdir(parents=True)
    with (packages_path / "test.py").open("w") as file:
        file.write("""X = 1""")
    yield truss_dir_path


@pytest.fixture
def custom_model_truss_dir_with_pre_and_post_str_example(tmp_path):
    dir_path = tmp_path / "custom_truss_with_pre_post_str_example"
    handle = init(str(dir_path))
    handle.spec.model_class_filepath.write_text(
        CUSTOM_MODEL_CODE_WITH_PRE_AND_POST_PROCESS
    )
    handle.update_examples(
        [
            Example(
                "example1",
                {
                    "inputs": [
                        {
                            "image_url": "https://github.com/pytorch/hub/raw/master/images/dog.jpg"
                        },
                    ],
                },
            )
        ]
    )
    yield dir_path


@pytest.fixture
def custom_model_truss_dir_with_pre_and_post_description(tmp_path):
    dir_path = tmp_path / "custom_truss_with_pre_post"
    handle = init(str(dir_path))
    handle.spec.model_class_filepath.write_text(
        CUSTOM_MODEL_CODE_WITH_PRE_AND_POST_PROCESS
    )
    handle.update_description("This model adds 3 to all inputs")
    yield dir_path


@pytest.fixture
def custom_model_truss_dir_for_gpu(tmp_path):
    dir_path = tmp_path / "custom_truss"
    handle = init(str(dir_path))
    handle.enable_gpu()
    handle.spec.model_class_filepath.write_text(CUSTOM_MODEL_CODE_FOR_GPU_TESTING)
    yield dir_path


@pytest.fixture
def custom_model_truss_dir_for_secrets(tmp_path):
    dir_path = tmp_path / "custom_truss"
    handle = init(str(dir_path))
    handle.add_secret("secret_name", "default_secret_value")
    handle.spec.model_class_filepath.write_text(CUSTOM_MODEL_CODE_FOR_SECRETS_TESTING)
    yield dir_path


@pytest.fixture
def truss_container_fs(tmp_path):
    ROOT = Path(__file__).parent.parent.parent.resolve()
    truss_fs = tmp_path / "truss_fs"
    truss_fs.mkdir()
    truss_build_dir = tmp_path / "truss_fs_build"
    truss_build_dir.mkdir()
    image_builder = ServingImageBuilderContext.run(
        ROOT / "truss" / "test_data" / "test_truss",
    )
    image_builder.prepare_image_build_dir(truss_build_dir)
    dockerfile_path = truss_build_dir / "Dockerfile"

    docker_build_emulator = DockerBuildEmulator(dockerfile_path, truss_build_dir)
    docker_build_emulator.run(truss_fs)
    return truss_fs


@pytest.fixture
def patch_ping_test_server():
    port = "5001"
    proc = subprocess.Popen(
        [
            "poetry",
            "run",
            "flask",
            "--app",
            "app",
            "run",
            "-p",
            port,
            "--host",
            "0.0.0.0",
        ],
        cwd=str(Path(__file__).parent.parent / "test_data" / "patch_ping_test_server"),
    )
    base_url = f"http://127.0.0.1:{port}"
    retry_secs = 10
    sleep_between_retries = 1
    for _ in range(int(retry_secs / sleep_between_retries)):
        time.sleep(sleep_between_retries)
        try:
            resp = requests.get(f"{base_url}/health")
        except requests.exceptions.ConnectionError:
            continue
        if resp.status_code == 200:
            break

    yield port
    proc.terminate()


def _pytorch_model_from_content(
    path: Path,
    content: str,
    model_module_name: str = "my_model",
    model_class_name: str = "MyModel",
    model_filename: str = "my_model.py",
):
    f = path / model_filename
    f.write_text(content)

    sys.path.append(str(path))
    model_class = getattr(importlib.import_module(model_module_name), model_class_name)
    return model_class(), f


def _custom_model_from_code(
    where_dir: Path,
    truss_name: str,
    model_code: str,
    handle_ops: callable = None,
) -> Path:
    dir_path = where_dir / truss_name
    handle = init(str(dir_path))
    if handle_ops is not None:
        handle_ops(handle)
    handle.spec.model_class_filepath.write_text(model_code)
    return dir_path


class Helpers:
    @staticmethod
    @contextlib.contextmanager
    def file_content(file_path: Path, content: str):
        orig_content = file_path.read_text()
        try:
            file_path.write_text(content)
            yield
        finally:
            file_path.write_text(orig_content)

    @staticmethod
    @contextlib.contextmanager
    def sys_path(path: Path):
        try:
            sys.path.append(str(path))
            yield
        finally:
            sys.path.pop()

    @staticmethod
    @contextlib.contextmanager
    def env_var(var: str, value: str):
        orig_environ = os.environ.copy()
        try:
            os.environ[var] = value
            yield
        finally:
            os.environ.clear()
            os.environ.update(orig_environ)


@pytest.fixture
def helpers():
    return Helpers()



import tempfile
from pathlib import Path

import numpy as np
import pytest
from truss.constants import CONFIG_FILE
from truss.contexts.local_loader.load_model_local import LoadModelLocal
from truss.model_frameworks.lightgbm import LightGBM
from truss.tests.test_testing_utilities_for_other_tests import ensure_kill_all
from truss.truss_config import TrussConfig
from truss.truss_handle import TrussHandle
from truss.types import ModelFrameworkType


def test_to_truss(lgb_pima_model):
    with tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, "truss")
        lgb_framework = LightGBM()
        lgb_framework.to_truss(lgb_pima_model, truss_dir)

        # Assertions
        config = TrussConfig.from_yaml(truss_dir / CONFIG_FILE)
        assert config.model_class_filename == "model.py"
        assert config.model_class_name == "Model"
        assert config.model_framework == ModelFrameworkType.LIGHTGBM
        assert config.model_type == "Model"
        assert config.python_version.startswith("py3")
        assert len(config.requirements) != 0

        model_metadata = config.model_metadata
        assert model_metadata["model_binary_dir"] == "model"
        # assert model_metadata['supports_predict_proba']

        assert (truss_dir / "data" / "model" / "model.joblib").exists()
        assert (truss_dir / "model" / "model.py").exists()


def test_run_truss(lgb_pima_model):
    with tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, "truss")
        lgb_framework = LightGBM()
        lgb_framework.to_truss(lgb_pima_model, truss_dir)
        model = LoadModelLocal.run(truss_dir)
        predictions = model.predict([[0, 0, 0, 0, 0, 0, 0, 0]])
        assert len(predictions["predictions"]) == 1
        assert len(predictions["predictions"][0]) == 2


@pytest.mark.integration
def test_run_image(lgb_pima_model):
    with ensure_kill_all(), tempfile.TemporaryDirectory(dir=".") as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, "truss")
        lgb_framework = LightGBM()
        lgb_framework.to_truss(lgb_pima_model, truss_dir)
        tr = TrussHandle(truss_dir)
        predictions = tr.docker_predict(
            [[0, 0, 0, 0, 0, 0, 0, 0]],
            local_port=8080,
        )
        assert len(predictions["predictions"]) == 1
        assert np.shape(predictions["predictions"]) == (1, 2)



from typing import Any

import xgboost as xgb

MODEL_BASENAME = "model"
MODEL_EXTENSIONS = [".ubj"]


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs["data_dir"]
        config = kwargs["config"]
        model_metadata = config["model_metadata"]
        # XGBoost models, saved and loaded via the native save/load
        # in XGBoost do not support predicting probabilities unless
        # they are a multi-class classification problem.
        # TODO: Integrate model_metadata field to determine model
        # objective function to determine if an XGBoost model
        # supports predicting probabilities.
        self._supports_predict_proba = False
        self._model_binary_dir = model_metadata["model_binary_dir"]
        self._model = None

    def load(self):
        model_binary_dir_path = self._data_dir / self._model_binary_dir
        paths = [
            (model_binary_dir_path / MODEL_BASENAME).with_suffix(model_extension)
            for model_extension in MODEL_EXTENSIONS
        ]
        model_file_path = next(path for path in paths if path.exists())
        self._model = xgb.Booster()
        self._model.load_model(model_file_path)

    def predict(self, model_input: Any) -> Any:
        model_output = {}
        dmatrix_inputs = xgb.DMatrix(model_input)
        result = self._model.predict(dmatrix_inputs)
        model_output["predictions"] = result
        return model_output



from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List

import dockerfile
from truss.util.path import copy_tree_or_file


@dataclass
class DockerBuildEmulatorResult:
    workdir: Path = field(default_factory=lambda: Path("/"))
    env: Dict = field(default_factory=dict)
    entrypoint: List = field(default_factory=list)


class DockerBuildEmulator:
    """Emulates Docker Builds

    As running docker builds is expensive, this class emulates the docker build
    by parsing the docker file and applying certain commands to create an
    appropriate enviroment in a directory to simulate the root of the file system.

    Support COPY, ENV, ENTRYPOINT, WORKDIR commands. All other commands are ignored.
    """

    def __init__(
        self,
        dockerfile_path: Path,
        context_dir: Path,
    ) -> None:
        self._commands = dockerfile.parse_file(str(dockerfile_path))
        self._context_dir = context_dir

    def run(self, fs_root_dir: Path) -> DockerBuildEmulatorResult:
        def _resolve_env(key: str) -> str:
            if key.startswith("$"):
                key = key.replace("$", "", 1)
                v = result.env[key]
                return v
            return key

        def _resolve_values(keys: List[str]) -> List[str]:
            return list(map(_resolve_env, keys))

        result = DockerBuildEmulatorResult()
        for cmd in self._commands:
            if cmd.cmd not in ["ENV", "ENTRYPOINT", "COPY", "WORKDIR"]:
                continue
            values = _resolve_values(cmd.value)
            if cmd.cmd == "ENV":
                result.env[values[0]] = values[1]
            if cmd.cmd == "ENTRYPOINT":
                result.entrypoint = list(values)
            if cmd.cmd == "COPY":
                src, dst = values
                src = src.replace("./", "", 1)
                dst = dst.replace("/", "", 1)
                copy_tree_or_file(self._context_dir / src, fs_root_dir / dst)
            if cmd.cmd == "WORKDIR":
                result.workdir = result.workdir / values[0]
        return result



from typing import Any

MODEL_BASENAME = "model"
MODEL_EXTENSIONS = [".joblib", ".pkl", ".pickle"]


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs["data_dir"]
        config = kwargs["config"]
        model_metadata = config["model_metadata"]
        # LightGBM does not implement a `predict_proba` function
        self._supports_predict_proba = False
        self._model_binary_dir = model_metadata["model_binary_dir"]
        self._model = None

    def load(self):
        import joblib

        model_binary_dir_path = self._data_dir / self._model_binary_dir
        paths = [
            (model_binary_dir_path / MODEL_BASENAME).with_suffix(model_extension)
            for model_extension in MODEL_EXTENSIONS
        ]
        model_file_path = next(path for path in paths if path.exists())
        self._model = joblib.load(model_file_path)

    def predict(self, model_input: Any) -> Any:
        model_output = {}
        result = self._model.predict(model_input)
        model_output["predictions"] = result
        if self._supports_predict_proba:
            model_output["probabilities"] = self._model.predict_proba(
                model_input
            ).tolist()
        return model_output



from pathlib import Path
from typing import Any, Callable, List, Optional

from truss.patch.calc_patch import calc_truss_patch
from truss.patch.signature import calc_truss_signature
from truss.templates.control.control.helpers.types import (
    Action,
    ModelCodePatch,
    Patch,
    PatchType,
    PythonRequirementPatch,
    SystemPackagePatch,
)
from truss.truss_config import TrussConfig


def test_calc_truss_patch_unsupported(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)

    # Unsupported directory should result in no patches
    (custom_model_truss_dir / "data").touch()
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)
    assert len(patches) == 0

    # Changes under unsupported directory should return None to reflect
    # inability to calculate patch.
    (custom_model_truss_dir / "data" / "dummy").touch()
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)
    assert patches is None


def test_calc_truss_patch_add_file(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    with (custom_model_truss_dir / "model" / "dummy").open("w") as file:
        file.write("content")
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.ADD,
            path="dummy",
            content="content",
        ),
    )


def test_calc_truss_patch_add_under_new_directory(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    new_dir = custom_model_truss_dir / "model" / "dir"
    new_dir.mkdir()
    (new_dir / "dummy").touch()
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.ADD,
            path="dir/dummy",
            content="",
        ),
    )


def test_calc_truss_patch_remove_file(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    (custom_model_truss_dir / "model" / "model.py").unlink()
    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.REMOVE,
            path="model.py",
        ),
    )


def test_calc_truss_patch_update_file(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    new_model_file_content = """
    class Model:
        pass
    """
    with (custom_model_truss_dir / "model" / "model.py").open("w") as model_file:
        model_file.write(new_model_file_content)

    patches = calc_truss_patch(custom_model_truss_dir, prev_sign)

    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.UPDATE, path="model.py", content=new_model_file_content
        ),
    )


def test_calc_truss_ignore_pycache(custom_model_truss_dir: Path):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    top_pycache_path = custom_model_truss_dir / "__pycache__"
    top_pycache_path.mkdir()
    (top_pycache_path / "bla.pyc").touch()
    model_pycache_path = custom_model_truss_dir / "model" / "__pycache__"
    model_pycache_path.mkdir()
    (model_pycache_path / "foo.pyo").touch()

    patches = calc_truss_patch(
        custom_model_truss_dir,
        prev_sign,
    )
    assert len(patches) == 0


def test_calc_truss_ignore_pycache_existing(custom_model_truss_dir: Path):
    # If __pycache__ existed before and there are no changes, there should be no
    # patches.
    top_pycache_path = custom_model_truss_dir / "__pycache__"
    top_pycache_path.mkdir()
    (top_pycache_path / "bla.pyc").touch()
    model_pycache_path = custom_model_truss_dir / "model" / "__pycache__"
    model_pycache_path.mkdir()
    (model_pycache_path / "foo.pyo").touch()
    sign = calc_truss_signature(custom_model_truss_dir)
    patches = calc_truss_patch(
        custom_model_truss_dir,
        sign,
    )
    assert len(patches) == 0


def test_calc_truss_ignore_changes_outside_patch_relevant_dirs(
    custom_model_truss_dir: Path,
):
    prev_sign = calc_truss_signature(custom_model_truss_dir)
    top_pycache_path = custom_model_truss_dir / "__pycache__"
    top_pycache_path.mkdir()
    (top_pycache_path / "README.md").touch()
    git_dir = custom_model_truss_dir / ".git"
    git_dir.mkdir()
    (git_dir / "dummy").touch()

    patches = calc_truss_patch(
        custom_model_truss_dir,
        prev_sign,
    )
    assert len(patches) == 0

    # Removing should also be ignored
    new_sign = calc_truss_signature(custom_model_truss_dir)
    (git_dir / "dummy").unlink()
    patches = calc_truss_patch(
        custom_model_truss_dir,
        new_sign,
    )
    assert len(patches) == 0


def test_calc_config_patches_add_python_requirement(custom_model_truss_dir: Path):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        lambda config: config.requirements.append("requests==1.0.0"),
    )
    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(
            action=Action.ADD,
            requirement="requests==1.0.0",
        ),
    )


def test_calc_config_patches_remove_python_requirement(custom_model_truss_dir: Path):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=lambda config: config.requirements.append("requests==1.0.0"),
        config_op=lambda config: config.requirements.clear(),
    )
    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(
            action=Action.REMOVE,
            requirement="requests",
        ),
    )


def test_calc_config_patches_update_python_requirement(custom_model_truss_dir: Path):
    def update_requests_version(config: TrussConfig):
        config.requirements[0] = "requests==2.0.0"

    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=lambda config: config.requirements.append("requests==1.0.0"),
        config_op=update_requests_version,
    )
    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.PYTHON_REQUIREMENT,
        body=PythonRequirementPatch(
            action=Action.UPDATE,
            requirement="requests==2.0.0",
        ),
    )


def test_calc_config_patches_add_remove_and_update_python_requirement(
    custom_model_truss_dir: Path,
):
    def config_pre_op(config: TrussConfig):
        config.requirements = [
            "requests==1.0.0",
            "jinja==4.0.0",
        ]

    def config_op(config: TrussConfig):
        config.requirements = [
            "requests==2.0.0",
            "numpy>=1.8",
        ]

    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=config_pre_op,
        config_op=config_op,
    )
    assert len(patches) == 3
    patches.sort(key=lambda patch: patch.body.requirement)
    assert patches == [
        Patch(
            type=PatchType.PYTHON_REQUIREMENT,
            body=PythonRequirementPatch(
                action=Action.REMOVE,
                requirement="jinja",
            ),
        ),
        Patch(
            type=PatchType.PYTHON_REQUIREMENT,
            body=PythonRequirementPatch(
                action=Action.ADD,
                requirement="numpy>=1.8",
            ),
        ),
        Patch(
            type=PatchType.PYTHON_REQUIREMENT,
            body=PythonRequirementPatch(
                action=Action.UPDATE,
                requirement="requests==2.0.0",
            ),
        ),
    ]


def test_calc_config_patches_non_python_or_system_requirement_change(
    custom_model_truss_dir: Path,
):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_op=lambda config: config.environment_variables.update({"foo": "bar"}),
    )
    assert patches is None


def test_calc_config_patches_add_system_package(custom_model_truss_dir: Path):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        lambda config: config.system_packages.append("curl"),
    )
    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.SYSTEM_PACKAGE,
        body=SystemPackagePatch(
            action=Action.ADD,
            package="curl",
        ),
    )


def test_calc_config_patches_remove_system_package(custom_model_truss_dir: Path):
    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=lambda config: config.system_packages.append("curl"),
        config_op=lambda config: config.system_packages.clear(),
    )
    assert len(patches) == 1
    patch = patches[0]
    assert patch == Patch(
        type=PatchType.SYSTEM_PACKAGE,
        body=SystemPackagePatch(
            action=Action.REMOVE,
            package="curl",
        ),
    )


def test_calc_config_patches_add_and_remove_system_package(
    custom_model_truss_dir: Path,
):
    def config_pre_op(config: TrussConfig):
        config.system_packages = [
            "curl",
            "jq",
        ]

    def config_op(config: TrussConfig):
        config.system_packages = [
            "curl",
            "libsnd",
        ]

    patches = _apply_config_change_and_calc_patches(
        custom_model_truss_dir,
        config_pre_op=config_pre_op,
        config_op=config_op,
    )
    patches.sort(key=lambda patch: patch.body.package)
    assert patches == [
        Patch(
            type=PatchType.SYSTEM_PACKAGE,
            body=SystemPackagePatch(
                action=Action.REMOVE,
                package="jq",
            ),
        ),
        Patch(
            type=PatchType.SYSTEM_PACKAGE,
            body=SystemPackagePatch(
                action=Action.ADD,
                package="libsnd",
            ),
        ),
    ]


def _apply_config_change_and_calc_patches(
    custom_model_truss_dir: Path,
    config_op: Callable[[TrussConfig], Any],
    config_pre_op: Optional[Callable[[TrussConfig], Any]] = None,
) -> List[Patch]:
    def modify_config(op):
        config_path = custom_model_truss_dir / "config.yaml"
        config = TrussConfig.from_yaml(config_path)
        op(config)
        config.write_to_yaml_file(config_path)

    if config_pre_op is not None:
        modify_config(config_pre_op)

    prev_sign = calc_truss_signature(custom_model_truss_dir)
    modify_config(config_op)
    return calc_truss_patch(custom_model_truss_dir, prev_sign)



# flake8: noqa F401

from pathlib import Path

from single_source import get_version

__version__ = get_version(__name__, Path(__file__).parent.parent)


def version():
    return __version__


from truss.build import (
    create,
    create_from_mlflow_uri,
    from_directory,
    init,
    kill_all,
    load,
    mk_truss,
)



import os
from pathlib import Path
from threading import Thread
from typing import Union

from application import create_app
from helpers.inference_server_starter import inference_server_startup_flow
from waitress.server import BaseWSGIServer, MultiSocketServer

CONTROL_SERVER_PORT = int(os.environ.get("CONTROL_SERVER_PORT", "8080"))
INFERENCE_SERVER_PORT = int(os.environ.get("INFERENCE_SERVER_PORT", "8090"))
PYTHON_EXECUTABLE_LOOKUP_PATHS = [
    "/usr/local/bin/python",
    "/usr/local/bin/python3",
    "/usr/bin/python",
    "/usr/bin/python3",
]


def _identify_python_executable_path() -> str:
    for path in PYTHON_EXECUTABLE_LOOKUP_PATHS:
        if Path(path).exists():
            return path

    raise RuntimeError("Unable to find python, make sure it's installed.")


if __name__ == "__main__":
    from waitress import create_server

    inf_serv_home: str = os.environ["APP_HOME"]
    python_executable_path: str = _identify_python_executable_path()
    application = create_app(
        {
            "inference_server_home": inf_serv_home,
            "inference_server_process_args": [
                python_executable_path,
                f"{inf_serv_home}/inference_server.py",
            ],
            "control_server_host": "*",
            "control_server_port": CONTROL_SERVER_PORT,
            "inference_server_port": INFERENCE_SERVER_PORT,
        }
    )

    # Perform inference server startup flow in background
    Thread(target=inference_server_startup_flow, args=(application,)).start()

    application.logger.info(
        f"Starting live reload server on port {CONTROL_SERVER_PORT}"
    )
    server: Union[BaseWSGIServer, MultiSocketServer] = create_server(
        application,
        host=application.config["control_server_host"],
        port=application.config["control_server_port"],
    )
    server.run()



from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional, Type, Union


class PatchType(Enum):
    """Types of console requests sent to Django and passed along to pynode."""

    MODEL_CODE = "model_code"
    PYTHON_REQUIREMENT = "python_requirement"
    SYSTEM_PACKAGE = "system_package"


class Action(Enum):
    """Types of console requests sent to Django and passed along to pynode."""

    ADD = "ADD"
    UPDATE = "UPDATE"
    REMOVE = "REMOVE"


@dataclass
class PatchBody:
    """Marker class"""

    action: Action

    @abstractmethod
    def to_dict(self):
        pass


@dataclass
class ModelCodePatch(PatchBody):
    path: str  # Relative to model module directory
    content: Optional[str] = None

    def to_dict(self):
        return {
            "action": self.action.value,
            "path": self.path,
            "content": self.content,
        }

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return ModelCodePatch(
            action=Action[action_str],
            path=patch_dict["path"],
            content=patch_dict["content"],
        )


@dataclass
class PythonRequirementPatch(PatchBody):
    # For uninstall this should just be the name of the package, but for update
    # should be the full line in the requirements.txt format.
    requirement: str

    def to_dict(self):
        return {
            "action": self.action.value,
            "requirement": self.requirement,
        }

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return PythonRequirementPatch(
            action=Action[action_str],
            requirement=patch_dict["requirement"],
        )


@dataclass
class SystemPackagePatch(PatchBody):
    # For uninstall this should just be the name of the package, but for update
    # should be the full line in the requirements.txt format.
    package: str

    def to_dict(self):
        return {
            "action": self.action.value,
            "package": self.package,
        }

    @staticmethod
    def from_dict(patch_dict: Dict):
        action_str = patch_dict["action"]
        return SystemPackagePatch(
            action=Action[action_str],
            package=patch_dict["package"],
        )


PATCH_BODY_BY_TYPE: Dict[
    PatchType,
    Type[Union[ModelCodePatch, PythonRequirementPatch, SystemPackagePatch]],
] = {
    PatchType.MODEL_CODE: ModelCodePatch,
    PatchType.PYTHON_REQUIREMENT: PythonRequirementPatch,
    PatchType.SYSTEM_PACKAGE: SystemPackagePatch,
}


@dataclass
class Patch:
    """Request to execute code on console."""

    type: PatchType
    body: PatchBody

    def to_dict(self):
        return {
            "type": self.type.value,
            "body": self.body.to_dict(),
        }

    @staticmethod
    def from_dict(patch_dict: Dict):
        typ = PatchType(patch_dict["type"])
        body = PATCH_BODY_BY_TYPE[typ].from_dict(patch_dict["body"])
        return Patch(typ, body)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional, Set

import yaml
from truss.constants import CONFIG_FILE, TEMPLATES_DIR
from truss.environment_inference.requirements_inference import infer_deps
from truss.model_inference import infer_python_version, map_to_supported_python_version
from truss.truss_config import DEFAULT_EXAMPLES_FILENAME, TrussConfig
from truss.types import ModelFrameworkType
from truss.util.path import copy_file_path, copy_tree_path


class ModelFramework(ABC):
    @abstractmethod
    def typ(self) -> ModelFrameworkType:
        pass

    @abstractmethod
    def required_python_depedencies(self) -> Set[str]:
        """Returns a set of packages required by this framework.

        e.g. {'tensorflow'}
        """
        pass

    def requirements_txt(self) -> List[str]:

        return list(infer_deps(must_include_deps=self.required_python_depedencies()))

    @abstractmethod
    def serialize_model_to_directory(self, model, target_directory: Path):
        pass

    @abstractmethod
    def model_metadata(self, model) -> Dict[str, str]:
        pass

    def model_type(self, model) -> str:
        return "Model"

    def model_name(self, model) -> Optional[str]:
        return None

    def to_truss(self, model, target_directory: Path) -> None:
        """Exports in-memory model to a Truss, in a target directory."""
        model_binary_dir = target_directory / "data" / "model"
        model_binary_dir.mkdir(parents=True, exist_ok=True)

        # Serialize model and write it
        self.serialize_model_to_directory(model, model_binary_dir)
        template_path = TEMPLATES_DIR / self.typ().value
        copy_tree_path(template_path / "model", target_directory / "model")
        examples_path = template_path / DEFAULT_EXAMPLES_FILENAME
        target_examples_path = target_directory / DEFAULT_EXAMPLES_FILENAME
        if examples_path.exists():
            copy_file_path(examples_path, target_examples_path)
        else:
            target_examples_path.touch()

        python_version = map_to_supported_python_version(infer_python_version())

        # Create config
        config = TrussConfig(
            model_name=self.model_name(model),
            model_type=self.model_type(model),
            model_framework=self.typ(),
            model_metadata=self.model_metadata(model),
            requirements=self.requirements_txt(),
            python_version=python_version,
        )
        with (target_directory / CONFIG_FILE).open("w") as config_file:
            yaml.dump(config.to_dict(), config_file)

    @abstractmethod
    def supports_model_class(self, model_class) -> bool:
        pass



import os
import time

from truss.util import path


def test_max_modified():
    epoch_time = int(time.time())
    with path.given_or_temporary_dir() as dir:
        time.sleep(0.1)
        t1 = path.get_max_modified_time_of_dir(dir)
        assert t1 > epoch_time
        time.sleep(0.1)
        os.makedirs(os.path.join(dir, "test"))
        t2 = path.get_max_modified_time_of_dir(dir)
        assert t2 > t1


test_max_modified()



import os
import random
import string
import tempfile
from contextlib import contextmanager
from distutils.dir_util import copy_tree, remove_tree
from distutils.file_util import copy_file
from pathlib import Path
from typing import List, Optional, Tuple, Union


def copy_tree_path(src: Path, dest: Path) -> List[str]:
    return copy_tree(str(src), str(dest))


def copy_file_path(src: Path, dest: Path) -> Tuple[str, str]:
    return copy_file(str(src), str(dest))


def copy_tree_or_file(src: Path, dest: Path) -> Union[List[str], Tuple[str, str]]:
    if src.is_file():
        return copy_file_path(src, dest)

    return copy_tree_path(src, dest)


def remove_tree_path(target: Path) -> None:
    return remove_tree(str(target))


def get_max_modified_time_of_dir(path: Path) -> float:
    max_modified_time = os.path.getmtime(path)
    for root, dirs, files in os.walk(path):
        if os.path.islink(root):
            raise ValueError(f"Symlinks not allowed in Truss: {root}")
        files = [f for f in files if not f.startswith(".")]
        dirs[:] = [d for d in dirs if not d.startswith(".")]
        max_modified_time = max(max_modified_time, os.path.getmtime(root))
        for file in files:
            max_modified_time = max(
                max_modified_time, os.path.getmtime(os.path.join(root, file))
            )
    return max_modified_time


@contextmanager
def given_or_temporary_dir(given_dir: Optional[Path] = None):
    if given_dir is not None:
        yield given_dir
    else:
        with tempfile.TemporaryDirectory() as temp_dir:
            yield Path(temp_dir)


def build_truss_target_directory(stub: str) -> Path:
    """Builds a directory under ~/.truss/models for the purpose of creating a Truss at."""
    rand_suffix = "".join(random.choices(string.ascii_uppercase + string.digits, k=6))
    target_directory_path = Path(
        Path.home(), ".truss", "models", f"{stub}-{rand_suffix}"
    )
    target_directory_path.mkdir(parents=True)
    return target_directory_path



class Error(Exception):
    """Base Truss Error"""

    def __init__(self, message: str):
        super(Error, self).__init__(message)
        self.message = message
        self.type = type


class PatchApplicatonError(Error):
    pass


class UnsupportedPatch(PatchApplicatonError):
    """Patch unsupported by this truss"""

    pass


class PatchFailedRecoverable(PatchApplicatonError):
    """Patch admissible but failed to apply. Recoverable via further patching."""

    pass


class PatchFailedUnrecoverable(PatchApplicatonError):
    """Patch admissible but failed to apply, leaving truss in unrecoverable state.
    Full deploy is required."""

    pass


class InadmissiblePatch(PatchApplicatonError):
    """Patch does not apply to current state of Truss."""

    pass


class ModelNotReady(Error):
    """Model has started running, but not ready yet."""

    pass



import sys
from pathlib import Path
from unittest.mock import Mock

import pytest

# Needed to simulate the set up on the model docker container
sys.path.append(
    str(
        Path(__file__).parent.parent.parent.parent.parent.parent
        / "templates"
        / "control"
        / "control"
    )
)

# Have to use imports in this form, otherwise isinstance checks fail on helper classes
from helpers.patch_applier import PatchApplier  # noqa
from helpers.types import Action, ModelCodePatch, Patch, PatchType  # noqa


@pytest.fixture
def patch_applier(truss_container_fs):
    return PatchApplier(truss_container_fs / "app", Mock())


def test_patch_applier_add(patch_applier: PatchApplier, truss_container_fs):
    patch = Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.ADD,
            path="dummy",
            content="",
        ),
    )
    patch_applier.apply_patch(patch)
    assert (truss_container_fs / "app" / "model" / "dummy").exists()


def test_patch_applier_remove(patch_applier: PatchApplier, truss_container_fs):
    patch = Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.REMOVE,
            path="model.py",
        ),
    )
    assert (truss_container_fs / "app" / "model" / "model.py").exists()
    patch_applier.apply_patch(patch)
    assert not (truss_container_fs / "app" / "model" / "model.py").exists()


def test_patch_applier_update(patch_applier: PatchApplier, truss_container_fs):
    new_model_file_content = """
    class Model:
        pass
    """
    patch = Patch(
        type=PatchType.MODEL_CODE,
        body=ModelCodePatch(
            action=Action.UPDATE,
            path="model.py",
            content=new_model_file_content,
        ),
    )
    patch_applier.apply_patch(patch)
    with (truss_container_fs / "app" / "model" / "model.py").open() as model_file:
        assert model_file.read() == new_model_file_content



from pathlib import Path
from typing import Any, Dict, Set

from truss.constants import MLFLOW_REQ_MODULE_NAMES
from truss.model_framework import ModelFramework
from truss.templates.server.common.util import model_supports_predict_proba
from truss.truss_handle import TrussHandle
from truss.types import ModelFrameworkType

MODEL_FILENAME = "model.joblib"


class Mlflow(ModelFramework):
    def typ(self) -> ModelFrameworkType:
        return ModelFrameworkType.MLFLOW

    def required_python_depedencies(self) -> Set[str]:
        return MLFLOW_REQ_MODULE_NAMES

    def serialize_model_to_directory(self, model, target_directory: Path):
        import mlflow

        if isinstance(model, str):
            self._download_model_from_uri(uri=model, target_directory=target_directory)
        elif isinstance(model, mlflow.pyfunc.PyFuncModel):
            self._download_model_from_pyfunc(
                model=model, target_directory=target_directory
            )

    def model_metadata(self, model) -> Dict[str, Any]:
        supports_predict_proba = model_supports_predict_proba(model)
        return {
            "model_binary_dir": "model",
            "supports_predict_proba": supports_predict_proba,
        }

    def supports_model_class(self, model_class) -> bool:
        model_framework, _, _ = model_class.__module__.partition(".")
        return model_framework == ModelFrameworkType.MLFLOW.value

    def to_truss(self, model, target_directory: Path) -> None:
        super().to_truss(model, target_directory)
        self._add_mlflow_requirements(str(target_directory))

    def _download_model_from_uri(self, uri: str, target_directory: Path):
        from mlflow.artifacts import download_artifacts

        download_artifacts(artifact_uri=uri, dst_path=target_directory)

    def _download_model_from_pyfunc(self, model, target_directory: Path):
        from mlflow.artifacts import download_artifacts

        run_id = model._model_meta.run_id
        download_artifacts(run_id=run_id, dst_path=target_directory)

    def _add_mlflow_requirements(self, target_directory: str):
        truss = TrussHandle(truss_dir=Path(target_directory))
        requirements_file = (
            truss._spec.data_dir / "model" / "model" / "requirements.txt"
        )
        if not requirements_file.exists():
            return
        truss.update_requirements_from_file(str(requirements_file))



import tempfile
from contextlib import contextmanager
from pathlib import Path

import numpy as np
from truss.build import cleanup, create, init
from truss.truss_spec import TrussSpec


def test_truss_init(tmp_path):
    dir_name = str(tmp_path)
    init(dir_name)
    spec = TrussSpec(Path(dir_name))
    assert spec.model_module_dir.exists()
    assert spec.data_dir.exists()
    assert spec.truss_dir == tmp_path
    assert spec.config_path.exists()


def test_truss_init_with_data_file_and_requirements_file_and_bundled_packages(
    tmp_path,
):
    dir_path = tmp_path / "truss"
    dir_name = str(dir_path)

    # Init data files
    data_path = tmp_path / "data.txt"
    with data_path.open("w") as data_file:
        data_file.write("test")

    # Init requirements file
    req_file_path = tmp_path / "requirements.txt"
    requirements = [
        "tensorflow==2.3.1",
        "uvicorn==0.12.2",
    ]
    with req_file_path.open("w") as req_file:
        for req in requirements:
            req_file.write(f"{req}\n")

    # init bundled packages
    packages_path = tmp_path / "dep_pkg"
    packages_path.mkdir()
    packages_path_file_py = packages_path / "file.py"
    packages_path_init_py = packages_path / "__init__.py"
    pkg_files = [packages_path_init_py, packages_path_file_py]
    for pkg_file in pkg_files:
        with pkg_file.open("w") as fh:
            fh.write("test")

    init(
        dir_name,
        data_files=[str(data_path)],
        requirements_file=str(req_file_path),
        bundled_packages=[str(packages_path)],
    )
    spec = TrussSpec(Path(dir_name))
    assert spec.model_module_dir.exists()
    assert spec.truss_dir == dir_path
    assert spec.config_path.exists()
    assert spec.data_dir.exists()
    assert spec.bundled_packages_dir.exists()
    assert (spec.data_dir / "data.txt").exists()
    assert spec.requirements == requirements
    assert (spec.bundled_packages_dir / "dep_pkg" / "__init__.py").exists()
    assert (spec.bundled_packages_dir / "dep_pkg" / "file.py").exists()


def test_create(sklearn_rfc_model, tmp_path):
    dir_path = tmp_path / "truss"
    data_file_path = tmp_path / "data.txt"
    with data_file_path.open("w") as data_file:
        data_file.write("test")
    req_file_path = tmp_path / "requirements.txt"
    requirements = [
        "tensorflow==2.3.1",
        "uvicorn==0.12.2",
    ]
    with req_file_path.open("w") as req_file:
        for req in requirements:
            req_file.write(f"{req}\n")
    scaf = create(
        sklearn_rfc_model,
        target_directory=dir_path,
        data_files=[str(data_file_path)],
        requirements_file=str(req_file_path),
    )
    spec = scaf.spec
    assert spec.model_module_dir.exists()
    assert spec.truss_dir == dir_path
    assert spec.config_path.exists()
    assert spec.data_dir.exists()
    assert (spec.data_dir / "data.txt").exists()
    assert spec.requirements == requirements


def test_create_pipeline(sklearn_rfc_model, tmp_path):
    def inference(model_input: dict):
        model_output = sklearn_rfc_model.predict([model_input])[0]
        return {"result": model_output}

    dir_path = tmp_path / "truss"
    data_file_path = tmp_path / "data.txt"
    with data_file_path.open("w") as data_file:
        data_file.write("test")
    req_file_path = tmp_path / "requirements.txt"
    requirements = [
        "tensorflow==2.3.1",
        "uvicorn==0.12.2",
    ]
    with req_file_path.open("w") as req_file:
        for req in requirements:
            req_file.write(f"{req}\n")
    scaf = create(
        inference,
        target_directory=dir_path,
        data_files=[str(data_file_path)],
        requirements_file=str(req_file_path),
    )
    spec = scaf.spec
    assert spec.model_module_dir.exists()
    assert spec.truss_dir == dir_path
    assert spec.config_path.exists()
    assert spec.data_dir.exists()
    assert (spec.data_dir / "data.txt").exists()
    assert spec.requirements == requirements


def test_truss_sklearn_predict(sklearn_rfc_model):
    with _model_server_predict(sklearn_rfc_model, [[0, 0, 0, 0]]) as result:
        assert "predictions" in result
        assert "probabilities" in result
        probabilities = result["probabilities"]
        assert np.shape(probabilities) == (1, 3)


def test_truss_sklearn_predict_pipeline(sklearn_rfc_model):
    def inference(model_input: dict):
        model_output = sklearn_rfc_model.predict([model_input])[0]
        return {"result": model_output}

    with _model_server_predict_pipeline(inference, [0, 0, 0, 0]) as result:
        assert "result" in result
        assert result["result"] == 0


def test_truss_keras_predict(keras_mpg_model):
    with _model_server_predict(
        keras_mpg_model,
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
    ) as result:
        assert "predictions" in result
        predictions = result["predictions"]
        assert np.shape(predictions) == (1, 1)


def test_truss_keras_predict_pipeline(keras_mpg_model):
    def inference(model_input: dict):
        model_output = keras_mpg_model.predict(model_input)
        return {"result": model_output}

    with _model_server_predict_pipeline(
        inference,
        [0, 0, 0, 0, 0, 0, 0, 0, 0],
    ) as result:
        assert "result" in result
        predictions = result["result"]
        assert np.shape(predictions) == (1, 1)


def test_truss_pytorch_predict(pytorch_model):
    model = pytorch_model[0]
    with _model_server_predict(
        model,
        [[0, 0, 0]],
    ) as result:
        assert "predictions" in result
        assert len(result["predictions"]) == 1


def test_truss_huggingface_transformer_predict(
    huggingface_transformer_t5_small_pipeline,
):
    with _model_server_predict(
        huggingface_transformer_t5_small_pipeline,
        "My name is Sarah and I live in London",
    ) as result:
        assert "predictions" in result
        predictions = result["predictions"]
        assert len(predictions) == 1
        assert predictions[0]["generated_text"].startswith("Mein Name")


def test_cleanup(sklearn_rfc_model, tmp_path):
    data_file_path = tmp_path / "data.txt"
    with data_file_path.open("w") as data_file:
        data_file.write("test")
    req_file_path = tmp_path / "requirements.txt"
    requirements = [
        "tensorflow==2.3.1",
        "uvicorn==0.12.2",
    ]
    with req_file_path.open("w") as req_file:
        for req in requirements:
            req_file.write(f"{req}\n")
    _ = create(
        sklearn_rfc_model,
        data_files=[str(data_file_path)],
        requirements_file=str(req_file_path),
    )
    cleanup()
    build_folder_path = Path(Path.home(), ".truss")
    directory = list(build_folder_path.glob("**/*"))
    files = [obj.name for obj in directory if obj.is_file()]
    unique_files = set(files)
    assert build_folder_path.exists()
    assert unique_files == {"config.yaml"}


def test_truss_via_simple_mk_pipeline():
    def generate(request):
        x = request["x"]
        return {"y": x + 1}

    with _model_server_predict_pipeline(generate, {"x": 1}) as result:
        assert result["y"] == 2


def test_truss_via_t5_mk_pipeline(
    huggingface_transformer_t5_small_model, huggingface_transformer_t5_small_tokenizer
):
    def generate(request):
        prompt = request["prompt"]
        input_ids = huggingface_transformer_t5_small_tokenizer(
            prompt, return_tensors="pt"
        ).input_ids
        return {
            "response": huggingface_transformer_t5_small_tokenizer.decode(
                huggingface_transformer_t5_small_model.generate(input_ids)[0],
                skip_special_tokens=True,
            )
        }

    with _model_server_predict_pipeline(
        generate, {"prompt": "translate to french: hello"}
    ) as result:
        assert result["response"].startswith("Hallo")


@contextmanager
def _model_server_predict(model, model_input):
    with tempfile.TemporaryDirectory() as dir_name:
        sc = create(model, target_directory=dir_name)
        result = sc.server_predict(model_input)
        yield result


@contextmanager
def _model_server_predict_pipeline(pipeline, model_input):
    sc = create(pipeline)
    result = sc.server_predict(model_input)
    yield result
